'use strict';

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _entries = require('babel-runtime/core-js/object/entries');

var _entries2 = _interopRequireDefault(_entries);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _layouter = require('../src/layouter.js');

var _layouter2 = _interopRequireDefault(_layouter);

var _imagesMock = require('./images-mock.js');

var _range = require('lodash/range');

var _range2 = _interopRequireDefault(_range);

var _chai = require('chai');

var _deepFreeze = require('deep-freeze');

var _deepFreeze2 = _interopRequireDefault(_deepFreeze);

var _expectedOffsets = require('./expectedOffsets');

var _expectedOffsets2 = _interopRequireDefault(_expectedOffsets);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var getItems = function getItems(count) {
  return (0, _deepFreeze2.default)(_imagesMock.testImages.slice(0, count));
};
var getGroupCount = function getGroupCount(layout) {
  return layout.columns.reduce(function (count, column) {
    return count + column.length;
  }, 0);
};

describe('Layouter', function () {

  var gallery = void 0;
  var layouter = void 0;
  var container = {};
  var styleParams = {};
  var styleParamsOptions = {};

  var getLayout = function getLayout(layoutParams) {
    return layouter.createLayout(layoutParams);
  };

  beforeEach(function () {

    var items = getItems();
    styleParams = {
      oneRow: false,
      isVertical: false,
      gallerySize: 200,
      groupSize: 3,
      groupTypes: '1,2h,2v,3t,3b,3l,3r,3v,3h',
      rotatingGroupTypes: '',
      cubeImages: false,
      cubeType: 'fill',
      cubeRatio: 1,
      smartCrop: false,
      chooseBestGroup: true,
      collageAmount: 0.9,
      collageDensity: 0.9,
      minItemSize: 20,
      layoutsVersion: 2,
      galleryMargin: 0,
      imageMargin: 10,
      floatingImages: 0,
      fixedColumns: 0
    };

    styleParamsOptions = {
      oneRow: [true, false],
      isVertical: [true, false],
      gallerySize: [100, 200, 500],
      groupSize: [1, 2, 3],
      groupTypes: ['1,2h,2v,3t,3b,3l,3r,3v,3h', '1,3t,3l,3h', '2h,2v,3t,3v,3h'],
      rotatingGroupTypes: ['', '1,2v,3t'],
      cubeImages: [true, false],
      cubeRatio: [1, 2, 0.5],
      smartCrop: [true, false],
      chooseBestGroup: [true, false],
      collageDensity: [0, 0.6, 1],
      minItemSize: [20, 50, 100],
      imageMargin: [0, 10, 50],
      fixedColumns: [0, 1, 5]
    };

    container = {
      galleryWidth: 1000,
      bounds: {
        visibleTop: 0,
        visibleBottom: 1000,
        renderedTop: 0,
        renderedBottom: 10000
      }
    };

    layouter = new _layouter2.default({ items: items, container: container, styleParams: styleParams });
  });

  describe('items', function () {
    it('should not change default layout', function () {

      var items = getItems(100);

      var res = {};

      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = (0, _getIterator3.default)((0, _entries2.default)(styleParamsOptions)), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var option = _step.value;

          var _option = (0, _slicedToArray3.default)(option, 2),
              styleParam = _option[0],
              values = _option[1];

          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = (0, _getIterator3.default)(values), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var value = _step2.value;

              styleParams[styleParam] = value;
              gallery = getLayout({ items: items, container: container, styleParams: styleParams });

              var offsets = gallery.items.reduce(function (str, item) {
                var os = item.offset;
                var itemStr = '{t:' + Math.round(os.top) + ',l:' + Math.round(os.left) + ',r:' + Math.round(os.right) + ',b:' + Math.round(os.bottom) + '}';
                return str + itemStr;
              }, '');

              var spStr = (styleParam + '_' + value).replace(/[,.]/g, '');

              res[spStr] = spStr + offsets;
              (0, _chai.expect)(_expectedOffsets2.default[spStr]).to.equal(spStr + offsets);
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }
        }

        // Use these lines to print an expected result
        // for (const a of Object.entries(res)) {
        //   console.log(`${a[0]}: ${a[1]}`);
        // }
        // expect(true).to.be.false;

      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    });

    it('should include all items in original order', function () {

      styleParams.galleryWidth = 500;
      styleParams.minItemSize = 160;

      var _arr = [10, 50, 100];
      for (var _i = 0; _i < _arr.length; _i++) {
        var size = _arr[_i];
        var items = getItems(size);
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var galleryItemIds = gallery.items.map(function (item) {
          return item.id;
        });
        var itemIds = items.map(function (item) {
          return item.id || item.photoId;
        });

        (0, _chai.expect)(galleryItemIds).to.deep.equal(itemIds);
      }
    });

    it('should have items offsets and dimensions calculated correctly', function () {

      var items = getItems(100);
      styleParams.galleryWidth = 4000;
      styleParams.gallerySize = 500;
      styleParams.rotatingGroupTypes = '1,2h,2v,3r,3t,3l,3b,3v,3h';
      styleParams.imageMargin = 0;

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });

      gallery.groups.forEach(function (group) {
        group.items.forEach(function (item) {
          var widthDiff = Math.abs(item.offset.right - item.offset.left - item.width);
          var heightDiff = Math.abs(item.offset.bottom - item.offset.top - item.height);
          (0, _chai.expect)(widthDiff).to.be.below(1);
          (0, _chai.expect)(heightDiff).to.be.below(1);
        });
      });
    });
  });

  describe('container', function () {
    it('should support legacy container props', function () {
      var items = getItems();

      var containerA = { galleryWidth: 100, galleryHeight: 200 };
      var layouterA = new _layouter2.default({ items: items, container: containerA, styleParams: styleParams });

      var containerB = { width: 100, height: 200 };
      var layouterB = new _layouter2.default({ items: items, container: containerB, styleParams: styleParams });

      (0, _chai.expect)(layouterA.container).to.deep.equal(layouterB.container);
    });
  });

  describe('Style Params', function () {

    //gallerySize
    it('should have taller Strips as gallerySize increases', function () {
      var items = getItems(100);
      styleParams.isVertical = false;

      var lastGroupHeight = 0;
      var _arr2 = [100, 200, 300, 400];
      for (var _i2 = 0; _i2 < _arr2.length; _i2++) {
        var size = _arr2[_i2];
        styleParams.gallerySize = size;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });
        var maxGroupHeight = gallery.columns[0].groups.reduce(function (g, group) {
          var isLastStrip = group.stripIdx === gallery.strips.length;
          return Math.max(g, isLastStrip ? 0 : group.height);
        }, 0);

        (0, _chai.expect)(maxGroupHeight).to.be.above(lastGroupHeight);

        lastGroupHeight = maxGroupHeight;
      }
    });

    //gallerySize
    it('should have all groups in a Columns gallery narrower than gallerySize', function () {
      var items = getItems(100);
      styleParams.isVertical = true;
      styleParams.galleryWidth = 1200;

      var lastGroupWidth = 0;
      var _arr3 = [10, 50, 100, 200, 300, 400];
      for (var _i3 = 0; _i3 < _arr3.length; _i3++) {
        var size = _arr3[_i3];
        styleParams.gallerySize = size;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var maxGroupWidth = gallery.columns.reduce(function (c, column) {
          return c && column.groups.reduce(function (g, group) {
            return Math.max(g, group.width);
          }, true);
        }, true);

        (0, _chai.expect)(maxGroupWidth).to.be.above(lastGroupWidth);

        lastGroupWidth = maxGroupWidth;
      }
    });

    //cubeRatio
    it('should have all images in a grid gallery in the required ratio', function () {

      var allowedRounding = 2; //the number of pixels that can change due to rounding

      var items = getItems(100);
      styleParams.cubeImages = true;
      styleParams.imageMargin = 0;

      var _arr4 = [0.25, 0.5, 1, 2, 4];
      for (var _i4 = 0; _i4 < _arr4.length; _i4++) {
        var ratio = _arr4[_i4];
        styleParams.cubeRatio = ratio;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var isCroppedCorrectly = gallery.columns[0].groups.reduce(function (g, group) {
          return g && group.isLastGroup || group.items.reduce(function (i, image) {
            //ignore the last group (it is stretched to fill the galleryWidth)
            var isItemCroppedCorrectly = (image.width - allowedRounding) / (image.height + allowedRounding) <= image.cropRatio && (image.width + allowedRounding) / (image.height - allowedRounding) >= image.cropRatio;
            return i && isItemCroppedCorrectly;
          }, true);
        }, true);

        (0, _chai.expect)(isCroppedCorrectly).to.be.true;
      }
    });

    //fixedColumns
    it('should have fixed number of columns if specified', function () {
      var items = getItems(100);
      styleParams.isVertical = true;

      var _arr5 = [1, 5, 10, 20];
      for (var _i5 = 0; _i5 < _arr5.length; _i5++) {
        var num = _arr5[_i5];
        styleParams.fixedColumns = num;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        (0, _chai.expect)(gallery.columns.length).to.equal(num);
      }
    });

    //collageAmount
    it('should have more items in groups when the collageAmount increases', function () {
      var itemCount = 100;
      var items = getItems(itemCount);
      var collageAmounts = (0, _range2.default)(0, 11).map(function (i) {
        return i / 10;
      });
      var lastGroupCount = itemCount;
      styleParams.layoutsVersion = 1;

      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = (0, _getIterator3.default)(collageAmounts), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var collageAmount = _step3.value;

          styleParams.collageAmount = collageAmount;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });
          var groupCount = getGroupCount(gallery);

          (0, _chai.expect)(groupCount).not.to.be.above(lastGroupCount);

          lastGroupCount = groupCount;
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    });

    //collageDensity
    it('should have more items in groups when the collageDensity increases', function () {
      var itemCount = 100;
      var items = getItems(itemCount);

      var collageDensities = (0, _range2.default)(0, 11).map(function (i) {
        return i / 10;
      });
      var lastGroupCount = itemCount;

      var _iteratorNormalCompletion4 = true;
      var _didIteratorError4 = false;
      var _iteratorError4 = undefined;

      try {
        for (var _iterator4 = (0, _getIterator3.default)(collageDensities), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
          var collageDensity = _step4.value;

          styleParams.collageDensity = collageDensity;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });
          var groupCount = getGroupCount(gallery);

          (0, _chai.expect)(groupCount).not.to.be.above(lastGroupCount);

          lastGroupCount = groupCount;
        }
      } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion4 && _iterator4.return) {
            _iterator4.return();
          }
        } finally {
          if (_didIteratorError4) {
            throw _iteratorError4;
          }
        }
      }
    });

    //groupSize
    it('should have all groups at maximum groupSize items', function () {
      var items = getItems(100);

      var _arr6 = [1, 2, 3];
      for (var _i6 = 0; _i6 < _arr6.length; _i6++) {
        var size = _arr6[_i6];
        styleParams.groupSize = size;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var isWithinSize = gallery.columns[0].groups.reduce(function (g, group) {
          var inSize = group.items.length <= styleParams.groupSize;
          return g && inSize;
        }, true);

        (0, _chai.expect)(isWithinSize).to.be.true;
      }
    });

    //groupTypes
    it('should have groups only from the optional groups types ', function () {
      var items = getItems(100);

      var groupTypes = ['1', '1,2h,2v', '1,3b,3l,3r', '1,2h,2v,3v,3h', '1,3t,3b', '1,3v,3h', '1,3r,3b,3v,3h', '1,2h,2v,3v,3h,3l,3b']; //groupType '1' must always be an option

      var _iteratorNormalCompletion5 = true;
      var _didIteratorError5 = false;
      var _iteratorError5 = undefined;

      try {
        for (var _iterator5 = (0, _getIterator3.default)(groupTypes), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
          var type = _step5.value;

          styleParams.groupTypes = type;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });

          var isWithinTypes = gallery.columns[0].groups.reduce(function (g, group) {
            var inTypes = styleParams.groupTypes.indexOf(group.type) >= 0;
            return g && inTypes;
          }, true);

          (0, _chai.expect)(isWithinTypes).to.be.true;
        }
      } catch (err) {
        _didIteratorError5 = true;
        _iteratorError5 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion5 && _iterator5.return) {
            _iterator5.return();
          }
        } finally {
          if (_didIteratorError5) {
            throw _iteratorError5;
          }
        }
      }
    });

    //minItemSize
    it('should have all Strips GalleryLayout images larger than minItemSize', function () {

      var items = getItems(100);
      styleParams.isVertical = false;
      styleParams.imageMargin = 0;

      var minItemSizes = [10, 50, 100, 200, 300, 400];

      var _iteratorNormalCompletion6 = true;
      var _didIteratorError6 = false;
      var _iteratorError6 = undefined;

      try {
        for (var _iterator6 = (0, _getIterator3.default)(minItemSizes), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
          var size = _step6.value;

          styleParams.gallerySize = size * 4; //gallerySize must be greater than minItemSize (otherwise the images' proportions will affect the minDimension)
          styleParams.minItemSize = size;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });

          var minItemSize = gallery.columns[0].groups.reduce(function (g, group) {
            return group.items.reduce(function (i, item) {
              var minDimension = Math.min(item.width, item.height);
              return Math.min(i, minDimension);
            }, styleParams.minItemSize);
          }, styleParams.minItemSize);

          (0, _chai.expect)(minItemSize).not.to.be.below(size);
        }
      } catch (err) {
        _didIteratorError6 = true;
        _iteratorError6 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion6 && _iterator6.return) {
            _iterator6.return();
          }
        } finally {
          if (_didIteratorError6) {
            throw _iteratorError6;
          }
        }
      }
    });

    //minItemSize
    it('should have all Columns GalleryLayout images larger than minItemSize', function () {

      var items = getItems(100);
      styleParams.isVertical = true;
      styleParams.galleryWidth = 4000;
      styleParams.imageMargin = 0;
      styleParams.collageDensity = 1;

      var minItemSizes = [10, 50, 100, 200, 300];

      var _iteratorNormalCompletion7 = true;
      var _didIteratorError7 = false;
      var _iteratorError7 = undefined;

      try {
        for (var _iterator7 = (0, _getIterator3.default)(minItemSizes), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
          var size = _step7.value;

          styleParams.gallerySize = size * 8; //gallerySize must be greater than minItemSize (otherwise the images' proportions will affect the minDimension)
          styleParams.minItemSize = size;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });

          var minItemSize = gallery.columns.reduce(function (c, column) {
            return c && column.groups.reduce(function (g, group) {
              return Math.min(g, group.items.reduce(function (i, item) {
                var maxDimension = Math.max(item.width, item.height);
                return Math.min(i, maxDimension);
              }, styleParams.minItemSize));
            }, styleParams.minItemSize);
          }, styleParams.minItemSize);

          (0, _chai.expect)(minItemSize).not.to.be.below(size);
        }
      } catch (err) {
        _didIteratorError7 = true;
        _iteratorError7 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion7 && _iterator7.return) {
            _iterator7.return();
          }
        } finally {
          if (_didIteratorError7) {
            throw _iteratorError7;
          }
        }
      }
    });

    // isVertical
    it('should create vertical layouts if isVertical is true', function () {

      var items = getItems(100);
      styleParams.gallerySize = 200;
      styleParams.fixedColumns = 0;
      container.galleryWidth = 1000;

      styleParams.isVertical = true;
      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      (0, _chai.expect)(gallery.columns.length).to.equal(5);

      styleParams.isVertical = false;
      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      (0, _chai.expect)(gallery.columns.length).to.equal(1);
    });

    // oneRow
    it('should create one long row of items if oneRow is true', function () {

      var items = getItems(100);
      container.galleryHeight = 500;

      styleParams.oneRow = false;
      styleParams.imageMargin = 10;

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      (0, _chai.expect)(gallery.height).to.be.above(container.galleryHeight);

      styleParams.oneRow = true;
      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      (0, _chai.expect)(gallery.height).to.equal(container.galleryHeight);
    });

    // rotatingGroupTypes
    it('should have groups from the rotating groups types by their order ', function () {
      var items = getItems(100);
      styleParams.isVertical = false;

      var groupTypes = ['1', '1,2h,2v', '1,3b,1,3r', '1,2h,2v,3v,3h', '1,3t,3b', '1,3v,3h', '1,3r,2h,3v,3h', '2h,2v,3v,3h,3l,3b'];

      var _iteratorNormalCompletion8 = true;
      var _didIteratorError8 = false;
      var _iteratorError8 = undefined;

      try {
        for (var _iterator8 = (0, _getIterator3.default)(groupTypes), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
          var type = _step8.value;

          styleParams.rotatingGroupTypes = type;
          gallery = getLayout({ items: items, container: container, styleParams: styleParams });

          var isWithinTypes = gallery.columns[0].groups.reduce(function (g, group, idx) {
            var rotatingGroupTypes = styleParams.rotatingGroupTypes.split(',');
            var expectedType = rotatingGroupTypes[idx % rotatingGroupTypes.length];
            var groupType = group.type;
            (0, _chai.expect)(expectedType).to.equal(groupType);
            var isType = expectedType === groupType;
            return g && isType;
          }, true);

          (0, _chai.expect)(isWithinTypes).to.be.true;
        }
      } catch (err) {
        _didIteratorError8 = true;
        _iteratorError8 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion8 && _iterator8.return) {
            _iterator8.return();
          }
        } finally {
          if (_didIteratorError8) {
            throw _iteratorError8;
          }
        }
      }
    });

    // cubeImages
    it('should have all images in their original ratio if cubeImages is false', function () {

      var allowedRounding = 2; //the number of pixels that can change due to rounding

      var items = getItems(100); //todo - something breaks when using exactly 100 images
      styleParams.cubeImages = false;
      styleParams.imageMargin = 0;
      styleParams.collageDensity = 0.8;

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      var isOriginalDimensions = gallery.columns[0].groups.reduce(function (g, group) {
        return g && group.isLastGroup || group.items.reduce(function (i, image) {
          //ignore the last group (it is stretched to fill the galleryWidth)
          var isItemCroppedCorrectly = (image.width - allowedRounding) / (image.height + allowedRounding) <= image.maxWidth / image.maxHeight && (image.width + allowedRounding) / (image.height - allowedRounding) >= image.maxWidth / image.maxHeight;
          return i && isItemCroppedCorrectly; //ignore fractions
        }, true);
      }, true);

      (0, _chai.expect)(isOriginalDimensions).to.be.true;

      styleParams.cubeImages = true;

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      var isCroppedCorrectly = gallery.columns[0].groups.reduce(function (g, group) {
        return g && group.items.reduce(function (i, image) {
          var isItemCroppedCorrectly = (image.width - allowedRounding) / (image.height + allowedRounding) <= styleParams.cubeRatio && (image.width + allowedRounding) / (image.height - allowedRounding) >= styleParams.cubeRatio;
          return i && isItemCroppedCorrectly;
        }, true);
      }, true);

      (0, _chai.expect)(isCroppedCorrectly).to.be.true;
    });

    // smartCrop
    it('should crop images according to their original proportions if smartCrop is true', function () {

      var allowedRounding = 2; //the number of pixels that can change due to rounding

      var items = getItems(100);
      styleParams.cubeRatio = 2;
      styleParams.cubeImages = true;
      styleParams.smartCrop = true;
      styleParams.imageMargin = 0;

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      var isCroppedCorrectly = gallery.columns[0].groups.reduce(function (g, group) {
        return g && group.items.reduce(function (i, image) {
          var cropRatio = image.isLandscape ? styleParams.cubeRatio : 1 / styleParams.cubeRatio;
          var isItemCroppedCorrectly = (image.width - allowedRounding) / (image.height + allowedRounding) <= cropRatio && (image.width + allowedRounding) / (image.height - allowedRounding) >= cropRatio;
          return i && isItemCroppedCorrectly;
        }, true);
      }, true);

      (0, _chai.expect)(isCroppedCorrectly).to.be.true;
    });

    // chooseBestGroup
    it('should not allow ugly groups if chooseBestGroup is true ', function () {
      var items = getItems(99);
      styleParams.groupTypes = '3t,3r,3l,3b'; //without 1
      styleParams.groupSize = 3;
      styleParams.collageDensity = 1;
      styleParams.minItemSize = 10;
      styleParams.gallerySize = 1000;

      var _arr7 = [true, false];
      for (var _i7 = 0; _i7 < _arr7.length; _i7++) {
        var chooseBestGroup = _arr7[_i7];
        styleParams.chooseBestGroup = chooseBestGroup;

        gallery = getLayout({ items: items, container: container, styleParams: styleParams });
        var isWithinTypes = gallery.columns[0].groups.reduce(function (g, group) {
          var groupType = group.type;
          var isType = groupType !== '1'; //if ChoooseBestGroup is true, some group will have to be single
          return g && isType;
        }, true);

        (0, _chai.expect)(isWithinTypes).to.not.equal(chooseBestGroup);
      }
    });

    // imageMargin (between groups)
    it('should have spaces between groups equal to imageMargin', function () {

      var items = getItems(100);
      styleParams.galleryWidth = 4000;
      styleParams.gallerySize = 500;
      styleParams.groupSize = 1;

      var _loop = function _loop(margin) {

        styleParams.imageMargin = margin;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var lastItem = false;
        var marginDiff = 0;
        var totalMarginDiff = gallery.items.reduce(function (i, item) {
          if (lastItem) {
            var realMargin = void 0;
            if (lastItem.offset.top === item.offset.top) {
              realMargin = Math.round(item.offset.left - lastItem.offset.right);
            } else {
              realMargin = Math.round(item.offset.top - lastItem.offset.bottom);
            }
            marginDiff = Math.abs(realMargin - margin * 2);
            (0, _chai.expect)(marginDiff).to.be.below(1);
          }
          lastItem = item;
          return i + Math.floor(marginDiff);
        }, 0);

        (0, _chai.expect)(totalMarginDiff).to.be.below(1);
      };

      var _arr8 = [10, 50, 100, 200];
      for (var _i8 = 0; _i8 < _arr8.length; _i8++) {
        var margin = _arr8[_i8];
        _loop(margin);
      }
    });

    // imageMargin (within groups)
    //TODO fix this test once the playground is complete
    it('should have spaces between items in a group equal to imageMargin', function () {

      var items = getItems(100);
      styleParams.galleryWidth = 4000;
      styleParams.gallerySize = 1000;
      styleParams.groupSize = 3;
      styleParams.groupTypes = '1,2h,2v,3r,3t,3l,3b,3v,3h';

      var _loop2 = function _loop2(margin) {

        styleParams.imageMargin = margin;
        gallery = getLayout({ items: items, container: container, styleParams: styleParams });

        var marginDiff = 0;
        gallery.columns[0].groups.reduce(function (g, group) {
          var lastItem = false;
          var groupMarginDiff = group.items.reduce(function (i, item) {
            if (lastItem) {
              var realMargin = void 0;
              if (lastItem.offset.top < item.offset.top) {
                realMargin = Math.round(item.offset.top - lastItem.offset.bottom);
              } else {
                realMargin = Math.round(item.offset.left - lastItem.offset.right);
              }
              marginDiff = Math.abs(realMargin - margin * 2);
              (0, _chai.expect)(marginDiff).to.be.below(2);
            }
            lastItem = item;
            return i + Math.floor(marginDiff);
          }, 0);
          return g + groupMarginDiff;
        }, 0);
      };

      var _arr9 = [0, 30, 40, 80];
      for (var _i9 = 0; _i9 < _arr9.length; _i9++) {
        var margin = _arr9[_i9];
        _loop2(margin);
      }
    });

    // rotatingGroupTypes
    it('should type groups according to rotatingGroupTypes if defined', function () {

      var items = getItems(100); //todo - something breaks when using exactly 100 images
      styleParams.rotatingGroupTypes = '2h,3v,3b,3t,1,2h,2v';
      var rotatingGroupTypesArr = styleParams.rotatingGroupTypes.split(',');

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      gallery.groups.forEach(function (group, g) {
        (0, _chai.expect)(group.type).to.equal(rotatingGroupTypesArr[(g - 1) % rotatingGroupTypesArr.length]); //first group idx is 1
      }, true);
    });

    // rotatingCropRatios
    it('should crop items according to rotatingCropRatios if defined', function () {

      var items = getItems(100); //todo - something breaks when using exactly 100 images
      styleParams.rotatingCropRatios = '0.2,0.6,8,4.5';
      styleParams.cropRatio = '1';
      styleParams.cropItems = true;
      styleParams.smartCrop = false;

      var rotatingCropRatiosArr = styleParams.rotatingCropRatios.split(',');

      gallery = getLayout({ items: items, container: container, styleParams: styleParams });
      gallery.items.forEach(function (item, i) {
        console.log(item);
        (0, _chai.expect)(item.cropRatio).to.equal(rotatingCropRatiosArr[i % rotatingCropRatiosArr.length]);
      }, true);
    });
  });

  describe('Infinite Scroll', function () {
    it('scroll should increase visible items range', function () {
      var items = getItems(100);

      styleParams.isVertical = true;
      styleParams.fixedColumns = 2;
      styleParams.groupSize = 1;

      container.galleryWidth = 1000;
      container.galleryHeight = 1000;

      var isVisible = function isVisible(itemIdx) {
        var item = gallery.items[itemIdx];
        return !!item && item.visibility.visible;
      };

      var visibleItemsRange = function visibleItemsRange(items) {
        var from = false;
        var to = false;
        var _iteratorNormalCompletion9 = true;
        var _didIteratorError9 = false;
        var _iteratorError9 = undefined;

        try {
          for (var _iterator9 = (0, _getIterator3.default)(items), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
            var item = _step9.value;

            if (isVisible(item.idx)) {
              if (from === false) {
                from = item.idx;
              }
            } else if (to === false && from !== false) {
              to = item.idx - 1;
            }
          }
        } catch (err) {
          _didIteratorError9 = true;
          _iteratorError9 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion9 && _iterator9.return) {
              _iterator9.return();
            }
          } finally {
            if (_didIteratorError9) {
              throw _iteratorError9;
            }
          }
        }

        return [from, to];
      };

      var scroll = function scroll(base) {
        return {
          renderedTop: base - 3000,
          visibleTop: base,
          visibleBottom: base + 1000,
          renderedBottom: base + 4000
        };
      };

      var lastFrom = -1;
      var lastTo = -1;

      var boundsOptions = [0, 1000, 2500, 6000].map(function (base) {
        return scroll(base);
      });
      gallery = layouter.createLayout({ items: items, container: container, styleParams: styleParams });

      var _iteratorNormalCompletion10 = true;
      var _didIteratorError10 = false;
      var _iteratorError10 = undefined;

      try {
        for (var _iterator10 = (0, _getIterator3.default)(boundsOptions), _step10; !(_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done); _iteratorNormalCompletion10 = true) {
          var bounds = _step10.value;


          gallery = layouter.calcVisibilities(bounds);

          var _visibleItemsRange = visibleItemsRange(gallery.items),
              _visibleItemsRange2 = (0, _slicedToArray3.default)(_visibleItemsRange, 2),
              from = _visibleItemsRange2[0],
              to = _visibleItemsRange2[1];

          (0, _chai.expect)(isVisible(from - 1)).to.be.false;
          (0, _chai.expect)(isVisible(from)).to.be.true;
          (0, _chai.expect)(isVisible(to)).to.be.true;
          (0, _chai.expect)(isVisible(to + 1)).to.be.false;

          (0, _chai.expect)(lastFrom).to.be.below(from);
          (0, _chai.expect)(lastTo).to.be.below(to);

          lastFrom = from;
          lastTo = to;
        }
      } catch (err) {
        _didIteratorError10 = true;
        _iteratorError10 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion10 && _iterator10.return) {
            _iterator10.return();
          }
        } finally {
          if (_didIteratorError10) {
            throw _iteratorError10;
          }
        }
      }
    });
  });

  describe('public methods', function () {
    it('findLastVisibleItemIdx should work', function () {
      var items = getItems(100);

      styleParams.isVertical = true;
      styleParams.fixedColumns = 1;
      styleParams.groupSize = 1;

      container.galleryWidth = 1000;

      var galleryHeights = [500, 1000, 2500, 6000];

      var _iteratorNormalCompletion11 = true;
      var _didIteratorError11 = false;
      var _iteratorError11 = undefined;

      try {
        for (var _iterator11 = (0, _getIterator3.default)(galleryHeights), _step11; !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
          var galleryHeight = _step11.value;


          container.galleryHeight = galleryHeight;
          gallery = layouter.createLayout({ items: items, container: container, styleParams: styleParams });

          var lastVisibleItemIdx = layouter.lastVisibleItemIdx();
          var lastVisibleItem = gallery.items[lastVisibleItemIdx];
          var firstHiddenItem = gallery.items[lastVisibleItemIdx + 1];

          (0, _chai.expect)(lastVisibleItem.offset.top).to.be.below(galleryHeight);
          (0, _chai.expect)(firstHiddenItem.offset.top).to.be.above(galleryHeight);
        }
      } catch (err) {
        _didIteratorError11 = true;
        _iteratorError11 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion11 && _iterator11.return) {
            _iterator11.return();
          }
        } finally {
          if (_didIteratorError11) {
            throw _iteratorError11;
          }
        }
      }
    });

    it('finfNeighborItem should work', function () {

      var neighbor = void 0;

      var items = getItems(100);

      styleParams.isVertical = true;
      styleParams.cubeImages = true;
      styleParams.cubeRatio = 1;
      styleParams.groupSize = 1;

      container.galleryWidth = 1000;

      var _arr10 = [2, 3, 5, 7];
      for (var _i10 = 0; _i10 < _arr10.length; _i10++) {
        var fixedColumns = _arr10[_i10];

        styleParams.fixedColumns = fixedColumns;
        gallery = layouter.createLayout({ items: items, container: container, styleParams: styleParams });

        var _iteratorNormalCompletion12 = true;
        var _didIteratorError12 = false;
        var _iteratorError12 = undefined;

        try {
          for (var _iterator12 = (0, _getIterator3.default)(gallery.items), _step12; !(_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done); _iteratorNormalCompletion12 = true) {
            var item = _step12.value;


            if (item.idx > fixedColumns) {
              neighbor = layouter.findNeighborItem(item.idx, 'up');
              (0, _chai.expect)(item.idx - neighbor).to.equal(fixedColumns);
            }

            if (item.idx < gallery.items.length - fixedColumns - 1) {
              neighbor = layouter.findNeighborItem(item.idx, 'down');
              (0, _chai.expect)(neighbor - item.idx).to.equal(fixedColumns);
            }

            if (item.idx % fixedColumns > 0) {
              neighbor = layouter.findNeighborItem(item.idx, 'left');
              (0, _chai.expect)(item.idx - neighbor).to.equal(1);
            }

            if (item.idx % fixedColumns < fixedColumns - 1 && item.idx < gallery.items.length - 1) {
              neighbor = layouter.findNeighborItem(item.idx, 'right');
              (0, _chai.expect)(neighbor - item.idx).to.equal(1);
            }
          }
        } catch (err) {
          _didIteratorError12 = true;
          _iteratorError12 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion12 && _iterator12.return) {
              _iterator12.return();
            }
          } finally {
            if (_didIteratorError12) {
              throw _iteratorError12;
            }
          }
        }
      }
    });
  });
});